#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/cdev.h>
#include <asm/system.h>		/* cli(), *_flags */
#include <asm/uaccess.h>	/* copy_*_user */
#include "scull.h"		/* local definitions */

/*
 * Our parameters which can be set at load time.
 */
int scull_major =   SCULL_MAJOR;		/* 豸 */
int scull_minor =   0;		/* 豸 */
int scull_nr_devs = SCULL_NR_DEVS;	/* number of bare scull devices */	
int scull_quantum = SCULL_QUANTUM;		/* ӴС(ֽ)ģװʱ */
int scull_qset =    SCULL_QSET;		/* ӼСģװʱ */

/* ģʱָĲ */
module_param(scull_major, int, S_IRUGO);
module_param(scull_minor, int, S_IRUGO);
module_param(scull_nr_devs, int, S_IRUGO);
module_param(scull_quantum, int, S_IRUGO);
module_param(scull_qset, int, S_IRUGO);

struct scull_dev *scull_devices;	/* allocated in scull_init_module */

/*
 * Empty out the scull device; must be called with the device
 * semaphore held.
 */
 /* 豸ļ */
 /* ͷ:Ӽ->Ӽ->ӡ򵥱ͷֵκӼ */
 /* scull_openļΪдʱ */
 /* øúʱҪź- */
int scull_trim(struct scull_dev *dev)
{
	struct scull_qset *next, *dptr;
	int qset = dev->qset;   /* devǿ */	/* ӼС,ӼӸָԪظ */	
	int i;

	/* 豸ӼѭΪ豸Ӽ */
	for (dptr = dev->data;	/* һӼ */
		dptr; 		/* dptrǷΪNULL */
		dptr = next) 		/* һӼ */
	{ 
		if (dptr->data) {		/* Ӽ(ָ)Ƿ */
			for (i = 0; i < qset; i++)	/* ͷŵǰӼеÿӣӼСΪqset */
				kfree(dptr->data[i]);		/* ͷһ(ָڴ)(ָڴ)СΪquantumֽ */
			kfree(dptr->data);		/* ͷһӼ(ָ)洢qset(ָ)ʱռݵڴ */
			dptr->data = NULL;
		}
		next = dptr->next;		/* ȡһӼ */
		kfree(dptr);		/* ͷŵǰӼ */
	}
	/* struct scull_dev *devбֵ */
	dev->size = 0;
	dev->quantum = scull_quantum;
	dev->qset = scull_qset;
	dev->data = NULL;
	return 0;
}
/*
 * Open and close
 */
/* ԭ:int (*open) (struct inode *, struct file *) */
int scull_open(struct inode *inode, struct file *filp)
{
	struct scull_dev *dev; /* device information */

	/* ʶҪ򿪵豸(õ豸Ӧ豸ṹ) */
	/* container_oføṹstruct scull_devĳԱstruct cdev cdevõָøṹָ */
	dev = container_of(inode->i_cdev, 		/* ָóԱָ룬struct inodeж壬struct cdev *i_cdev */
					struct scull_dev, 	/* ṹ */
					cdev);		/* óԱƣڸṹstruct scull_devУstruct cdev cdev */
	/* õdevṹ浽filp->private_dataÿ򿪵豸(ļ)һfileṹ壬filpָ */
	filp->private_data = dev; /* for other methods */	

	/* now trim to 0 the length of the device if open was write-only */
	/* ֻдʽ򿪣scull_trimض豸Ϊ0 */
	if ( (filp->f_flags & O_ACCMODE) == O_WRONLY) {		/* ж豸(ļ)Ĵģʽ */
		if (down_interruptible(&dev->sem))		/* Է */
			return -ERESTARTSYS;
		scull_trim(dev); /* ignore errors */	/* øúʱҪź- */
		up(&dev->sem);		/*  */
	}
	return 0;          /* success */
}

/* ԭ:int (*release) (struct inode *, struct file *) */
/* úopen෴ÿcloseöreleaseĵ */
int scull_release(struct inode *inode, struct file *filp)		
{
	return 0;
}
/*
 * Follow the list
 */
 /* dev豸ĵnӼָ룬Ӽnµ */
struct scull_qset *scull_follow(struct scull_dev *dev, int n)
{
	struct scull_qset *qs = dev->data;		/* ǰ豸ĵһӼ */

        /* Allocate first qset explicitly if need be */
        /* ǰ豸ûӼʾطһӼ */
	if (! qs) {
		/* kmallocַַ̬ڴռ䣬Сڴ */
		qs = dev->data = kmalloc(sizeof(struct scull_qset), 	/* ҪĿС */
								GFP_KERNEL);		/* ڴΪ־ */
		if (qs == NULL)
			return NULL;  /* Never mind */		/* ʧ */
		memset(qs, 0, sizeof(struct scull_qset));		/* ڴ */
	}

	/* Then follow the list */
	/* ǰ豸ӼnȷnӼӼµ */
	while (n--) {
		if (!qs->next) {	/* Ӽnµ */
			qs->next = kmalloc(sizeof(struct scull_qset), GFP_KERNEL);
			if (qs->next == NULL)		/* ʧ */
				return NULL;  /* Never mind */
			memset(qs->next, 0, sizeof(struct scull_qset));
		}
		qs = qs->next;
		continue;		/* ѭ */
	}
	return qs;		/* dev豸ĵnӼָ */
}

/*
 * Data management: read and write
 */
/* ԭ:ssize_t (*read) (struct file *, char __user *, size_t, loff_t *) */
/* scull_readÿֻһ */
ssize_t scull_read(struct file *filp, 		/* ļָ */
				char __user *buf, 		/* ûռ仺¶ */
				size_t count,		/* ȡݳ */
                		loff_t *f_pos)		/* ļǰ/дλãfilp˽еƫ */
{
	struct scull_dev *dev = filp->private_data; 	/* õ豸ͨfilp->private_dataѱе豸ṹָ */		
	struct scull_qset *dptr;	/* the first listitem */	/* һӼָ */
	int quantum = dev->quantum, qset = dev->qset;		/* ӡӼС */
	int itemsize = quantum * qset; /* how many bytes in the listitem */	/* һӼֽܵ */
	int item, s_pos, q_pos, rest;
	ssize_t retval = 0;

	if (down_interruptible(&dev->sem))		/* Է */
		return -ERESTARTSYS;
	if (*f_pos >= dev->size)		/* жȡλѾС */
		goto out;
	if (*f_pos + count > dev->size)
		count = dev->size - *f_pos;		/* ȡݵĳ */

	/* find listitem, qset index, and offset in the quantum */
	/* /ӼжλдλãڵڼӼеĵڼӣеƫ */
	item = (long)*f_pos / itemsize;		/* ڼӼ */
	rest = (long)*f_pos % itemsize;		/* Ӽеƫ */
	s_pos = rest / quantum; 		/* ڼ*/
	q_pos = rest % quantum;		/* еƫ */
	
	/* follow the list up to the right position (defined elsewhere) */
	/* õ豸devĵitemȡӼָ */
	dptr = scull_follow(dev, 		/* 豸ṹ */
					item);	/* ڼӼ */

	if (dptr == NULL || !dptr->data || ! dptr->data[s_pos])/* 豸ûӼӼûӡӸs_pos */
		goto out; /* don't fill holes */

	/* read only up to the end of this quantum */
	/* ֻһж:countǰӣضcount */
	if (count > quantum - q_pos)
		count = quantum - q_pos;

	/* ȡݲƵûռbufcountֽڣɹ0  */
	if (copy_to_user(buf, 		/* void __user *to */
			dptr->data[s_pos] + q_pos,		/* const void *from */
			count))		/* unsigned long n */
	{
		retval = -EFAULT;
		goto out;
	}
	*f_pos += count;		/* ļǰ/дλ */
	retval = count;		/* ʵʶȡݳ */

  out:
	up(&dev->sem);		/*  */
	return retval;
}

/* scull_writeÿֻһ */
ssize_t scull_write(struct file *filp, 		/* 豸Ӧļṹ */
					const char __user *buf, 		/* ûռ仺Ŵд */
					size_t count,		/* дݳ */
			                loff_t *f_pos)		/* ļǰ/дλãfilp˽еƫ */
{
	struct scull_dev *dev = filp->private_data;		/* ȡ豸ṹ */
	struct scull_qset *dptr;		/* һӼָ */
	int quantum = dev->quantum, qset = dev->qset;		/* ӡӼС */
	int itemsize = quantum * qset;		/* һӼֽܵ */
	int item, s_pos, q_pos, rest;
	ssize_t retval = -ENOMEM; /* value used in "goto out" statements */

	if (down_interruptible(&dev->sem))		/* Է */
		return -ERESTARTSYS;

	/* find listitem, qset index and offset in the quantum */
	/* /ӼжλдλãڵڼӼеĵڼӣеƫ */
	item = (long)*f_pos / itemsize;		/* ڼӼ */
	rest = (long)*f_pos % itemsize;		/* Ӽеƫ */
	s_pos = rest / quantum;		/* ڼ*/	
	q_pos = rest % quantum;		/* еƫ */

	/* follow the list up to the right position */
	/* õ豸devĵitemдݵӼָ */
	dptr = scull_follow(dev, item);
	if (dptr == NULL)		/* ȡʧ */
		goto out;
	if (!dptr->data) {		/* Ӽ(ָ)Ϊգӣ̬ */		
		dptr->data = kmalloc(qset * sizeof(char *), /* sizeof(char *):(ָ)ռĴСqset * sizeof(char *):ӼռڴС */
							GFP_KERNEL);
		if (!dptr->data)		/* ̬ʧ */
			goto out;
		memset(dptr->data, 0, qset * sizeof(char *));		/* ʼӼ */
	}
	if (!dptr->data[s_pos]) {		/* Ӽs_posһڴ */
		dptr->data[s_pos] = kmalloc(quantum, 		/* ӴСָڴС */
							GFP_KERNEL);	
		if (!dptr->data[s_pos])		/* ̬ʧ */
			goto out;
	}
	/* write only up to the end of this quantum */
	/* ֻڵǰд */
	if (count > quantum - q_pos)
		count = quantum - q_pos;		/* дݳ */

	/* ûռ俽ݵں˿ռ䣬ɹ0 */
	if (copy_from_user(dptr->data[s_pos]+q_pos, 		/* void *to */
					buf,		/* const void __user *from */
					count))		/* unsigned long n */
	{
		retval = -EFAULT;
		goto out;
	}
	*f_pos += count;		/* ļǰ/дλ */
	retval = count;		/* ʵдݳ */

        /* update the size */
	if (dev->size < *f_pos)
		dev->size = *f_pos;

  out:
	up(&dev->sem);		/*  */
	return retval;
}


struct file_operations scull_fops = {
	.owner =    THIS_MODULE,
	.read =     scull_read,
	.write =    scull_write,
	.open =     scull_open,		/* ڵַ */
	.release =  scull_release,
};/* עǻṹʼ﷨ */

/*
 * Finally, the module stuff
 */

/*
 * The cleanup function is used to handle initialization failures as well.
 * Thefore, it must be careful to work correctly even if some of the items
 * have not been initialized
 */
 /* ģ˳ */
void scull_cleanup_module(void)
{
	int i;
	dev_t devno = MKDEV(scull_major, scull_minor);		/* õscull豸豸 */

	/* Get rid of our char dev entries */
	/* ַ豸 */
	if (scull_devices) {
		for (i = 0; i < scull_nr_devs; i++) {
			scull_trim(scull_devices + i);		/* ͷ豸ṹ */
			cdev_del(&scull_devices[i].cdev);		/* Ƴcdev */				
		}
		kfree(scull_devices);		/* ͷ豸ṹscull_devices */
	}

	/* cleanup_module is never called if registering failed */
	/* עעscull_nr_devs豸 */
	unregister_chrdev_region(devno,		/* 豸עĵһ豸 */
						scull_nr_devs);			/* ע豸ŵĸ */

}


/*
 * Set up the char_dev structure for this device.
 */
 /* 豸עắ */
static void scull_setup_cdev(struct scull_dev *dev, int index)
{
	int err, devno = MKDEV(scull_major, scull_minor + index);		/* 豸ŵõ豸 */
														/* 豸:scull_major */							
    														/* 豸:scull_minor + index */
	cdev_init(&dev->cdev, &scull_fops);		/* ʼcdevṹָopsָ */
	dev->cdev.owner = THIS_MODULE;		/* ָcdevṹ */
	//dev->cdev.ops = &scull_fops;		/* һʡԣΪcdev_initʱʵ */
	/* ںע豸Ч */
	err = cdev_add (&dev->cdev,		/* 豸Ӧcdevṹ */ 
				devno,		/* 豸Ӧĵһ豸 */
				1);		/* ͸豸豸ŵĿȡ1ÿעһ豸 */
	/* Fail gracefully if need be */
	if (err)	/* ںע豸ʧ */
		printk(KERN_NOTICE "Error %d adding scull%d", err, index);
}

/* ģں */
int scull_init_module(void)
{
	int result, i;
	dev_t dev = 0;

/*
 * Get a range of minor numbers to work with, asking for a dynamic
 * major unless directed otherwise at load time.
 */
 	/* ܹںʶ豸 */
 	/* ̬豸alloc_chrdev_region */
 	/* ֪豸register_chrdev_region */
	if (scull_major) {
		dev = MKDEV(scull_major, scull_minor);		/* ֪豸ŵõ豸 */
		/* ʹ֪豸ܹںʶ */
		result = register_chrdev_region(dev,		/* Ҫ豸ŷΧʼֵ */ 
								scull_nr_devs,		/* 豸ŵĸ */
								"scull");		/* ñŷΧ豸 */
	} else {
		/* ̬豸 */
		result = alloc_chrdev_region(&dev, 		/* ڱѷ䷶Χĵһ豸 */
							scull_minor,		/* һ豸ţͨ0 */
							scull_nr_devs,		/* 豸ŵĸ */
							"scull");		/* ñŷΧ豸 */
		scull_major = MAJOR(dev);		/* Ӷ̬豸еõ豸 */
	}
	if (result < 0) {		/* ɹִУresult==0 */
		printk(KERN_WARNING "scull: can't get major %d\n", scull_major);
		return result;
	}

        /* 
	 * allocate the devices -- we can't have them static, as the number
	 * can be specified at load time
	 */
	 /* ʹkmalloc̬ظ4scull_devṹͨں˿ռڴ */
	scull_devices = kmalloc(scull_nr_devs * sizeof(struct scull_dev), GFP_KERNEL);
	if (!scull_devices) {		/* ڴʧ */
		result = -ENOMEM;
		goto fail;  /* Make this more graceful */	/* ʹgotoÿֵĴд */
	}
	/* ʼkmallocڴ */
	memset(scull_devices, 0, scull_nr_devs * sizeof(struct scull_dev));

        /* Initialize each device. */
        /* ʼÿ豸ķ--struct scull_devṹ */
	for (i = 0; i < scull_nr_devs; i++) {
		scull_devices[i].quantum = scull_quantum;		/* 趨ӴС */
		scull_devices[i].qset = scull_qset;				/* 趨ӼС */
		init_MUTEX(&scull_devices[i].sem);				/* 趨 */
		scull_setup_cdev(&scull_devices[i], i);			/* ںעַ豸 */		
	}

	return 0; /* succeed */

  fail:
  	/* scull_cleanup_moduleĴ˳ */
	scull_cleanup_module();
	return result;
}

module_init(scull_init_module);
module_exit(scull_cleanup_module);

MODULE_AUTHOR("ycz9999");
MODULE_LICENSE("Dual BSD/GPL");
